unit Snake;
// (c)2007 by Paul TOTH <tothpaul@free.fr>
// http://tothpaul.free.fr

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

uses
 Windows,OpenGL;

const
 GRID_SIZE  =16;  // taille de la grille
 SNAKE_TICKS=250; // temps en millisecondes pour passer d'une position  la suivante
 SNAKE_SPEED=20;  // dplacement entre deux cases

type
 TGrid=array[0..GRID_SIZE-1,0..GRID_SIZE-1] of byte;

 TSnake=class;

 TVector3f=record
  x,y,z:single
 end;

 TPosition=record
  x,y:integer;
 end;

 TSnakie=class
 private
  fOwner :TSnake;
  fParent:TSnakie;
  fNext  :TSnakie;
  fPos   :TVector3f;
  fMove  :TVector3f;
  fStep  :TVector3f;
  fTurn  :single;
  fAngle :single;
  fRotX  :single;
  fRotY  :single;
  fRotZ  :single;
  fTicks :cardinal;
  fCell  :TPosition;
  procedure HalfRender;
  function NextCell:TPosition;
 public
  constructor Create(AOwner:TSnake; AParent:TSnakie);
  destructor Destroy; override;
  procedure Animate(ticks:cardinal);
  procedure Render(Shadow:boolean);
  property Cell:TPosition read fCell;
  property Pos:TVector3f read fStep;
  property RotY:single read fRotY;
 end;

 TSnake=class
 private
  fRed  :integer;
  fGreen:integer;
  fHead :TSnakie;
  fLast :TSnakie;
  fMove :char;
  fGrid :TGrid;
  fDead :boolean;
  fDie  :cardinal;
  fFrog:TPosition;
  fScore:integer;
  function GetTexture(Color:integer):integer;
  procedure RedTexture;
  procedure GreenTexture;
 public
  procedure Reset;
  procedure Animate(ticks:cardinal);
  procedure Render(Shadow:boolean);
  procedure Grow;
  procedure RenderFrog;
  procedure MoveFrog;
  property Head:TSnakie read fHead;
  property Move:char write fMove;
  property Dead:boolean read fDead;
  property Die:cardinal read fDie;
  property Score:integer read fScore;
 end;

function FixAngle(Angle:single):single;

implementation

// ces deux fonctions manquent dans OpenGL.pas
procedure glGenTextures(count:integer; Handles:pointer); stdcall external OpenGL32;
procedure glBindTexture(Target,Texture:integer); stdcall external OpenGL32;

// ramener un angle entre 0 et 360
function FixAngle(Angle:single):single;
begin
 while Angle>360 do Angle:=Angle-360;
 while Angle<  0 do Angle:=Angle+360;
 Result:=Angle;
end;

// reset du serpent
procedure TSnake.Reset;
var
 i:integer;
begin
// supprimer tous les "snakies" (s'ils existent)
 fHead.Free;
// recrer la tte
 fHead:=TSnakie.Create(Self,nil);
// et les 4 premiers anneaux
 for i:=0 to 3 do TSnakie.Create(Self,fLast);
// on n'est pas mort
 fDead:=False;
// depuis pas de temps
 fDie:=0;
// la grille est vide
 FillChar(fGrid,SizeOf(fGrid),0);
// la grenouille est invisible
 fFrog.x:=-1;
// et le score n'est pas trs lev
 fScore:=0;
end;

// ajouter un anneau au serpent
procedure TSnake.Grow;
begin
 TSnakie.Create(Self,fLast);
// on gagne 1 point
 inc(fScore);
end;

// novelle grenouille
procedure TSnake.MoveFrog;
var
 x,y:integer;
begin
// si on est mort, on la fait juste sortir de la grille
 if fDead then begin
  if fFrog.x>=0 then fGrid[fFrog.x,fFrog.y]:=0;
  fFrog.x:=-1;
  exit;
 end;
// recherche une case libre
 repeat
  x:=Random(GRID_SIZE);
  y:=Random(GRID_SIZE);
 until fGrid[x,y]=0;
// si on avait une grenouille, la supprimer de la grille
 if fFrog.x>=0 then fGrid[fFrog.x,fFrog.y]:=0;
// placer la nouvelle
 fGrid[x,y]:=2;
 fFrog.X:=x;
 fFrog.Y:=y;
end;

// construire une texture colore avec un bord blanc
function TSnake.GetTexture(Color:integer):integer;
const
 k=16;
var
 bitmap:array of cardinal;
 i,j:integer;
begin
// faire un bitmap color
 SetLength(bitmap,k*k);
 for i:=0 to k-1 do begin
  for j:=0 to k-1 do begin
   bitmap[i+k*j]:=Color;
  end;
 end;
// ajouter le bord blanc
 for i:=0 to k-1 do begin
  bitmap[i      ]:=$ffffff;
  bitmap[k*i    ]:=$ffffff;
  bitmap[k*i+k-1]:=$ffffff;
  bitmap[k*k-i-1]:=$ffffff;
 end;
// en faire une texture OpenGL
 glGenTextures(1,@Result);
 glBindTexture(GL_TEXTURE_2D,Result);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
 glTexImage2D(GL_TEXTURE_2D, 0, 4, k,k,0, GL_RGBA, GL_UNSIGNED_BYTE,@bitmap[0]);
end;

// obtenir une texture Rouge ( bords blanc)
procedure TSnake.RedTexture;
begin
 if fRed=0 then fRed:=GetTexture($2010f0) else glBindTexture(GL_TEXTURE_2D,fRed);
end;

// obtenir une texture Verte ( bords blanc)
procedure TSnake.GreenTexture;
begin
 if fGreen=0 then fGreen:=GetTexture($20f010) else glBindTexture(GL_TEXTURE_2D,fGreen);
end;

// animer le serpent, s'il est en vie
procedure TSnake.Animate(Ticks:cardinal);
begin
 if fDead then
  inc(fDie,Ticks)
 else
  fHead.Animate(Ticks);
end;

// dessiner le serpent
procedure TSnake.Render(Shadow:boolean);
begin
 glEnable(GL_TEXTURE_2D);
// quand il meurt, le serpent s'enfonce dans le sol
 if fDead then glTranslatef(0,-fDie/200,0);
 fHead.Render(Shadow);
 glDisable(GL_TEXTURE_2D);
end;

// dessiner la grenouille
procedure TSnake.RenderFrog;
const
 k=4;
var
 i:integer;
begin
// sauf si elle est invisible
 if fFrog.x<0 then exit;
// c'est un cube vert
 GreenTexture;
 glEnable(GL_TEXTURE_2D);
 glPushMatrix;
  glTranslatef(20*(fFrog.x-8),0,20*(fFrog.y-8));
  glRotatef(45,0,1,0);
  for i:=0 to 4 do begin
   glBegin(GL_QUADS);
    glTexCoord2f(0,0); glVertex3f(-k,-k,+k);
    glTexCoord2f(1,0); glVertex3f(+k,-k,+k);
    glTexCoord2f(1,1); glVertex3f(+k,+k,+k);
    glTexCoord2f(0,1); glVertex3f(-k,+k,+k);
   glEnd;
   glRotatef(90,0,1,0);
   if i=3 then glRotatef(-90,1,0,0)
  end;
 glPopMatrix;
 glDisable(GL_TEXTURE_2D);
end;

//----------------------------------------------------------------------------//
// les anneaux sont des "snakie"

constructor TSnakie.Create(AOwner:TSnake; AParent:TSnakie);
begin
 fOwner:=AOwner;
 fOwner.fLast:=Self;
 fParent:=AParent;
 if fParent=nil then begin
 // positionner la tte dans la trape
  fMove.y:=SNAKE_SPEED;
  fPos.y:=-5*SNAKE_SPEED;
  fCell.x:=8;
  fCell.y:=8;
 end else begin
 // positionner chaque anneau en fonction du prcdent
  fParent.fNext:=Self;
  fPos :=fParent.fPos;
  fTicks:=fParent.fTicks;
  fCell:=fParent.fCell;
  fAngle:=fParent.fAngle;
  fStep:=fParent.fStep;
  fRotY:=fParent.fRotY;
  fRotZ:=fParent.fRotZ;
 end;
end;

// dtruire tous les anneaux
destructor TSnakie.Destroy;
begin
 if fNext<>nil then fNext.Free;
 inherited;
end;

// dessiner la moitier d'un anneau (voir Render)
procedure TSnakie.HalfRender;
const
 w1=3;
 w2=6;
 h1=6;
 h2=9;
 z =3;
begin
 glBegin(GL_QUADS);
  glTexCoord2f(0,0);  glVertex3f(+w1,+z,+h1);
  glTexCoord2f(1,0);  glVertex3f(-w1,+z,+h1);
  glTexCoord2f(1,1);  glVertex3f(-w2, 0,+h2);
  glTexCoord2f(0,1);  glVertex3f(+w2, 0,+h2);

  glTexCoord2f(0,0);  glVertex3f(-w1,+z,-h1);
  glTexCoord2f(1,0);  glVertex3f(+w1,+z,-h1);
  glTexCoord2f(1,1);  glVertex3f(+w2, 0,-h2);
  glTexCoord2f(0,1);  glVertex3f(-w2, 0,-h2);

  glTexCoord2f(0,0);  glVertex3f(+w1,+z,-h1);
  glTexCoord2f(1,0);  glVertex3f(-w1,+z,-h1);
  glTexCoord2f(1,1);  glVertex3f(-w1,+z,+h1);
  glTexCoord2f(0,1);  glVertex3f(+w1,+z,+h1);

  glTexCoord2f(0,0);  glVertex3f(-w2, 0,+h2);
  glTexCoord2f(1,0);  glVertex3f(-w1,+z,+h1);
  glTexCoord2f(1,1);  glVertex3f(-w1,+z,-h1);
  glTexCoord2f(0,1);  glVertex3f(-w2, 0,-h2);

  glTexCoord2f(0,0);  glVertex3f(+w2, 0,-h2);
  glTexCoord2f(1,0);  glVertex3f(+w1,+z,-h1);
  glTexCoord2f(1,1);  glVertex3f(+w1,+z,+h1);
  glTexCoord2f(0,1);  glVertex3f(+w2, 0,+h2);
 glEnd();
end;

// calcul de la prochaine position de l'anneau
function TSnakie.NextCell:TPosition;
begin
 Result:=fCell;
 if fMove.x>0 then inc(Result.x) else
 if fMove.x<0 then dec(Result.x) else
 if fMove.z>0 then inc(Result.y) else
 if fMove.z<0 then dec(Result.y);
end;

// animer un anneau
procedure TSnakie.Animate(ticks:cardinal);
var
 f:single;
begin
// commencer par bouger ceux qui suivent (car ils s'appuient sur la position actuelle de celui-ci)
 if fNext<>nil then fNext.Animate(ticks);
// compter le temps
 inc(fTicks,ticks);
// rattraper le temps perdu
 while fTicks>SNAKE_TICKS do begin
 // quand on est sorti de la trape
  if fMove.y=0 then begin
   // et que le serpent n'est pas mort
   if fOwner.fDead=False then begin
    // si on est le dernier, librer la cellule
    if fNext=nil then fOwner.fGrid[fCell.x,fCell.y]:=0;
    // passer  la suivante
    fCell:=NextCell;
    // la rserver
    fOwner.fGrid[fCell.x,fCell.y]:=1;
   end;
  end;
  // dplacer l'anneau
  fPos.x:=fPos.x+fMove.x;
  fPos.y:=fPos.y+fMove.y;
  fPos.z:=fPos.z+fMove.z;
  // appliquer la rotation
  fAngle:=FixAngle(fAngle+fTurn);
  // dplacement de la tte
  if fParent=nil then begin
  // quand on sort de la trape
   if fMove.Y>0 then begin
    // partir au nord
    if fPos.Y=0 then begin
     fMove.y:=0;
     fMove.z:=-SNAKE_SPEED;
    end;
   end else begin
   // tenir compte du clavier
    case fOwner.fMove of
     'L' : begin
      if fMove.x>0 then begin fMove.x:=0; fMove.z:=-SNAKE_SPEED end else
      if fMove.x<0 then begin fMove.x:=0; fMove.z:=+SNAKE_SPEED end else
      if fMove.z>0 then begin fMove.z:=0; fMove.x:=+SNAKE_SPEED end else
      if fMove.z<0 then begin fMove.z:=0; fMove.x:=-SNAKE_SPEED end;
      fTurn:=+90;
     end;
     'R' : begin
      if fMove.x>0 then begin fMove.x:=0; fMove.z:=+SNAKE_SPEED end else
      if fMove.x<0 then begin fMove.x:=0; fMove.z:=-SNAKE_SPEED end else
      if fMove.z>0 then begin fMove.z:=0; fMove.x:=-SNAKE_SPEED end else
      if fMove.z<0 then begin fMove.z:=0; fMove.x:=+SNAKE_SPEED end;
      fTurn:=-90;
     end;
     else fTurn:=0;
    end;
   end;
   fOwner.Move:=#0;
  end else begin
  // sinon suivre celui qui nous prcde
   fMove:=fParent.fMove;
   fTurn:=fParent.fTurn;
  end;
  dec(fTicks,SNAKE_TICKS);
 end;
 // proportionnel au temps qui passe
 f:=fTicks/SNAKE_TICKS;
 // au bout d'1/4 du mouvement, tester les collisions
 if (fParent=nil)and(f>1/4)and(fMove.Y=0) then begin
  with NextCell do begin
   if (X<0)or(Y<0)or(X>=GRID_SIZE)or(Y>=GRID_SIZE)or(fOwner.fGrid[x,y]=1) then fOwner.fDead:=True else
   if fOwner.fGrid[x,y]=2 then begin
    fOwner.Grow;
    fOwner.MoveFrog;
   end;
  end;
 end;
 // calculer la position courante
 fStep.x:=fPos.x+f*fMove.x;
 fStep.y:=fPos.y+f*fMove.y;
 fStep.z:=fPos.z+f*fMove.z;
 fRotY:=FixAngle(fAngle+f*fTurn);
 if f>1/2 then fRotZ:=(1-f)*fTurn else fRotZ:=f*fTurn;
 if fMove.Y>0 then begin
  f:=-fPos.y-f*fMove.y;
  if f>20 then f:=1 else f:=f/20;
  fRotX:=f*90;
 end else fRotX:=0;
end;

// dessiner un anneau
procedure TSnakie.Render(Shadow:boolean);
begin
// viter de dessiner le reflet au dpart, sinon il serait au dessus du sol
 if Shadow and (fMove.Y<>0) then exit;
 if fNext<>nil then fNext.Render(Shadow);
// choix de la couleur (tte rouge, corps vert)
 if fParent=nil then fOwner.RedTexture else fOwner.GreenTexture;
 glPushMatrix;
  glTranslatef(fStep.x,fStep.y,fStep.z);
  glRotatef(fRotY,0,1,0);
  glRotatef(fRotZ,0,0,1);
  glRotatef(fRotX,1,0,0);
 // dessiner la moitier suprieure
  if Shadow then glColor3f(0.35,0.35,0.35) else  glColor3f(1,1,1);
  HalfRender;
 // rotation  180 pour dessiner la moitier infrieure
  glRotatef(180,1,0,0);
  if Shadow then glColor3f(0.20,0.20,0.20) else glColor3f(0.5,0.5,0.5);
  HalfRender;
  if Shadow then glColor3f(0.35,0.35,0.35) else  glColor3f(1,1,1); // pour la grenouille
 glPopMatrix;
end;

end.

